#include "SimpleAnalysis/AnalysisClass.h"
#include <array>
#include <random>

/*-----------------------------------------------------------------------------
* HHMEtLowMass2016
* Particle-level implementation of low-mass analysis from SUSY-2017-02
* (Targets H~H~ --> hh+MEt --> bbbb+MEt)
* David Wardrope, 18/02/21
-----------------------------------------------------------------------------*/

DefineAnalysis(HHMEtLowMass2016)

std::mt19937 m_randomEngine;
float calcMEff(const float& met, const std::array<AnalysisObject, 4>& jets);
float dijetSelection(std::array<AnalysisObject, 4>& bjets);
bool passDEta(const std::array<AnalysisObject, 4>& bjets);
bool passMDC(const std::array<AnalysisObject, 4>& bjets);
bool passTopVeto(const std::array<AnalysisObject, 4>& hjets, const std::vector<AnalysisObject>& jets);
std::array<AnalysisObject, 4> selectFourJetsRandomly(AnalysisObjects&);

void HHMEtLowMass2016::Init()
{
	addRegions({"low_SR_MET0_meff440", "low_SR_MET150_meff440"});
	addHistogram("event_n", 1, 0, 1);
	addHistogram("cutflow1_n", 1, 0, 1);
	addHistogram("cutflow2_n", 1, 0, 1);
	addHistogram("cutflow3_n", 1, 0, 1);
	addHistogram("cutflow4_n", 1, 0, 1);
	addHistogram("cutflow5_n", 1, 0, 1);

	addHistogram("jet_n", 6, 4, 10);
	addHistogram("bjet_n", 6, 4, 10);
	addHistogram("bjet4_n", 6, 4, 10);
	addHistogram("m4j", 50, 250., 1250.);
	addHistogram("xhh", 50, 0., 10.);
	addHistogram("xhhPostTopVeto", 50, 0., 10.);
	addHistogram("lepton_n", 10, 0, 10);
	addHistogram("met", 50, 0, 1000);
	addHistogram("meff", 50, 0, 1000);
	addHistogram("meff_metGT150", 50, 0, 1000);
	std::random_device rd;
    m_randomEngine = std::mt19937(rd());

}
void HHMEtLowMass2016::ProcessEvent(AnalysisEvent* event)
{
	fill("event_n", 0.5);
	auto jets = event->getJets(40., 2.5);
	if(jets.size() < 4) return;
	//Overlap removal
	auto electrons  = event->getElectrons(5, 2.47, ELooseLH);
	// baseline muons are required to pass the Medium selections and to have pT > 5 GeV, |eta| < 2.5
	auto muons      = event->getMuons(5, 2.5, MuMedium);
	auto radiusCalcJet  = [] (const AnalysisObject& , const AnalysisObject& muon) { return std::min(0.4, 0.04 + 10/muon.Pt()); };
	auto radiusCalcMuon = [] (const AnalysisObject& muon, const AnalysisObject& ) { return std::min(0.4, 0.04 + 10/muon.Pt()); };
	auto radiusCalcElec = [] (const AnalysisObject& elec, const AnalysisObject& ) { return std::min(0.4, 0.04 + 10/elec.Pt()); };

	electrons  = overlapRemoval(electrons, muons, 0.01);
	jets   = overlapRemoval(jets, electrons, 0.2);
	electrons  = overlapRemoval(electrons, jets, radiusCalcElec);
	jets   = overlapRemoval(jets, muons, radiusCalcJet, LessThan3Tracks);
	muons      = overlapRemoval(muons, jets, radiusCalcMuon);

	if(jets.size() < 4) return;
	fill("cutflow1_n", 0.5);
	fill("jet_n", jets.size());
	auto bjets = filterObjects(jets, 40., 2.5, BTag70MV2c10);
	if(bjets.size() < 4) return;
	fill("cutflow2_n", 0.5);
	fill("bjet_n", bjets.size());
	//Select four b-jets at random from vector
	std::array<AnalysisObject, 4> hjets = selectFourJetsRandomly(bjets);
	fill("bjet4_n", hjets.size());

	float m4j = (hjets[0] + hjets[1] + hjets[2] + hjets[3]).M(); 
	fill("m4j", m4j);
	float xhh = dijetSelection(hjets);
	if(xhh < 0) return; //Negative value indicates suitable pair of Higgs candidates was not found.
	fill("xhh", xhh);
	if(!passTopVeto(hjets, jets)) return;
	fill("xhhPostTopVeto", xhh);

	//Lepton vetos.
	auto signalElectrons = filterObjects(electrons, 20, 2.47, ETightLH|ED0Sigma5|EZ05mm|EIsoBoosted);
	auto signalMuons     = filterObjects(muons, 20, 2.5, MuD0Sigma3|MuZ05mm|MuIsoBoosted|MuNotCosmic);
	unsigned int lepton_n = signalElectrons.size()+signalMuons.size();
	if(lepton_n > 0) return;
	fill("lepton_n", lepton_n);

	if(xhh > 1.6) return;
	fill("cutflow3_n", 0.5);

	//Now need Discovery Region definitions.
	float met = event->getGenMET();
	fill("met", met);
	float meff = calcMEff(met, hjets);
  	if(meff > 440)
	{
		fill("cutflow4_n", 0.5);
		fill("meff", meff);
		accept("low_SR_MET0_meff440");
  		if(met>150.)
		{
			fill("cutflow5_n", 0.5);
			fill("meff_metGT150", meff);
			accept("low_SR_MET150_meff440");
		}
	}
}
std::array<AnalysisObject, 4> selectFourJetsRandomly(AnalysisObjects& bjets)
{
	//Randomly shuffle jets and select the first four for Higgs candidate formation.
	//Use indices to save overhead?
	//std::vector<int> indices;
	//for(int i = 0; i < bjets.size(); ++i) indices.push_back(i);
	if(bjets.size() > 4)
	{
		std::shuffle(bjets.begin(), bjets.end(), m_randomEngine);	
	}
	//Can't use the following code, since no default constructor for AnalysisObject
	//std::array<AnalysisObject, 4> hjets;
	//for(int i = 0; i < bjets.size(); ++i) hjets[i] = bjets[i];
	std::array<AnalysisObject, 4> hjets = {bjets[0], bjets[1], bjets[2], bjets[3]};
	return hjets;
}
//dijetSelection tries to find Higgs boson candidates from an input of array<AnalysisObject>.
//It re-orders the jets in this array, such that bjets[0] and [1] correspond to Higgs 1 and [2] and [3] to Higgs 2
float dijetSelection(std::array<AnalysisObject, 4>& bjets)
{
	std::array<AnalysisObject, 4> input = bjets;
	//input jets are sorted by pt.

	float m4j = (input[0] + input[1] + input[2] + input[3]).M();
	//Consider three possible pairings of Higgs bosons into candidates:
	//Six combinations of labels "1" & "2", but actually a trivial relabelling reduces this to three.
	//Always pair (i, j) and (k, l)
	float bestDhh = 9.999e6;
	float bestDijetsXhh = -99;
	unsigned int i = 0;	
	for (unsigned int j=1; j< 4; ++j) 
	{
		//Now find the indices of the other two jets and form two dijets
		unsigned int k = 1; 
		unsigned int l = 1;
		for (; k < 4; ++k) 
		{
			if(k != i && k != j)
			{
				break;
			}
		}
		for (; l < 4; ++l) 
		{
			if(l != i && l != j && l != k)
			{
				break;
			}
		}

		//Sort dijets by scalar sum of jet pts.
		int j00, j01, j10, j11;
		if(input[k].Pt() + input[l].Pt() > input[i].Pt() + input[j].Pt())
		{
			j00 = k;
			j01 = l;
			j10 = i;
			j11 = j;
		}else
		{
			j00 = i;
			j01 = j;
			j10 = k;
			j11 = l;
		}
		std::vector<TLorentzVector> dijets;
		dijets.push_back(input[j00] + input[j01]);
		dijets.push_back(input[j10] + input[j11]);
		//Apply m4j-dependent DR(j, j) requirements
		float minDR0 = 0.;
		float maxDR0 = 1.;
		float minDR1 = 0.;
		float maxDR1 = 1.;
		if(m4j < 1250)
		{
			minDR0 = (360./m4j) - 0.5;	
			maxDR0 = (653./m4j) + 0.475;
			minDR1 = (235/m4j);
			maxDR1 = (875/m4j) + 0.35;
		}
		float dR0 = input[j00].DeltaR(input[j01]);
		if(dR0 < minDR0 || dR0 > maxDR0) continue;
		float dR1 = input[j10].DeltaR(input[j11]);
		if(dR1 < minDR1 || dR1 > maxDR1) continue;

		float dhh = fabs(dijets[0].M() - (120./110.)*dijets[1].M());
		if(dhh < bestDhh)	
		{	
			bestDhh = dhh;
			bjets[0] = input[j00];
			bjets[1] = input[j01];
			bjets[2] = input[j10];
			bjets[3] = input[j11];
			float xhh1 = (dijets[0].M() - 120)/(0.1*dijets[0].M());
			float xhh2 = (dijets[1].M() - 110)/(0.1*dijets[1].M());
			bestDijetsXhh = sqrt(xhh1*xhh1+xhh2*xhh2);
		}
	}

	return bestDijetsXhh;
}
//Following function is not used in this analysis.
bool passMDC(const std::array<AnalysisObject, 4>& bjets)
{
	float m4j = (bjets[0] + bjets[1] + bjets[2] + bjets[3]).M();
	std::vector<TLorentzVector> dijets;
	dijets.push_back(bjets[0] + bjets[1]);
	dijets.push_back(bjets[2] + bjets[3]);
	if(dijets[0].Pt() < 0.5*m4j - 90) return false;
	if(dijets[1].Pt() < 0.33*m4j - 70) return false;
	return true;
}

//Following function is not used in this analysis.
bool passDEta(const std::array<AnalysisObject, 4>& bjets)
{
	float maxDEta = 1.5;
	std::vector<TLorentzVector> dijets;
	dijets.push_back(bjets[0] + bjets[1]);
	dijets.push_back(bjets[2] + bjets[3]);

	if(fabs(dijets[0].Eta()-dijets[1].Eta()) > maxDEta) return false;
	return true;
}
bool passTopVeto(const std::array<AnalysisObject, 4>& hjets, const std::vector<AnalysisObject>& jets)
{
	//Form hadronic top candidates from at least one Higgs candidate constituent jet (makes sense, since they 
	//are b-tagged) and two other jets. One of these other jets must not be a Higgs constituent, to insist 
	//that it is a light jet. (Done in main analysis in terms of Higgs candidate constituents rather than simply
	//asking b-tagging because it makes more sense in context of background model construction.)
	for(auto bjet: hjets)
	{
		for(std::vector<AnalysisObject>::const_iterator wJet1 = jets.begin(); wJet1 < jets.end()-1; ++wJet1)
		{
			if(wJet1->index() == bjet.index()) continue;
			bool wJet1FromHiggs = false;
			for(auto hjet: hjets)
			{
				if(hjet.index() == wJet1->index()) wJet1FromHiggs = true;
			}
			for(std::vector<AnalysisObject>::const_iterator wJet2 = wJet1+1; wJet2 < jets.end(); ++wJet2)
			{
				if(wJet2->index() == bjet.index()) continue;
				bool wJet2FromHiggs = false;
				for(auto hjet: hjets)
				{
					if(hjet.index() == wJet2->index()) wJet2FromHiggs = true;
				}
				if(wJet1FromHiggs && wJet2FromHiggs) continue;

				TLorentzVector pW = (*wJet1)+(*wJet2);
				TLorentzVector pTop = bjet+pW;
				float xt = sqrt(std::pow((pW.M()-80.4)/(0.1*pW.M()), 2) + std::pow((pTop.M() - 172.5)/(0.1*pTop.M()), 2));
				if(xt < 1.8) return false;
			}
		}
	}		
	return true;
}
float calcMEff(const float& met, const std::array<AnalysisObject, 4>& jets)
{
	float meff = met;
	for(auto jet: jets)
	{
		meff += jet.Pt();
	}
	return meff;
}


